/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-events.h"
#include "x-image.h"
#include "fsel.h"
#include "graymap.h"
#include "chdir.h"
#include "btn.h"

#include "chdir-scr.c"

static int fc_butpos[]=
{ 45, 36, 176, 34 , 'O',
  10, 449, 65, 24 , 'C' };



typedef struct
{
  Window window; int w,h;
  ximage_t *image;
  int selbut[5];
  void (*func)(void *, char *);
  void *arg;
  file_selector_t *fs;
} fdir_t;

static void fc_destroy(fdir_t *fc, XEvent *ev);

static void fc_press(fdir_t *fc, XButtonEvent *ev)
{
  btn_click(fc->window, fc->image, fc_butpos, 10, ev, fc->selbut);
}

static void fc_release(fdir_t *fc, XButtonEvent *ev)
{
  Window window;
  switch(btn_release(fc->window, fc->image, ev, fc->selbut))
  {
  case 'O':
     if (fc->func) (*fc->func)(fc->arg, file_selector_directory(fc->fs));
  case 'C':
     window= fc->window;
     file_selector_free(fc->fs);
     fc_destroy(fc,NULL);
     XDestroyWindow(x_display, window);
     break;
  }
}

static void fc_destroy(fdir_t *fc, XEvent *ev)
{
  destroy_image(fc->image);
  unregister_event_handler(fc->window, Expose, NULL);
  unregister_event_handler(fc->window, ButtonPress, NULL);
  unregister_event_handler(fc->window, ButtonRelease, NULL);
  unregister_event_handler(fc->window, DestroyNotify, NULL);
  free(fc);
  printf("dir selector destroyed\n");
}


static void fc_expose(fdir_t *fc, XExposeEvent *ev)
{
  put_image(fc->window, ev->x, ev->y, fc->image,ev->x, ev->y, 
            ev->width, ev->height);
}

static void click_file(void *arg, file_selector_t *fs, char *sel)
{
  // IGNORED
}

void choose_dir(Window parent,
                 char *dir, void (*finish)(void *, char *), void *arg)
{
  XSetWindowAttributes attribs;
  fdir_t *fc;
  int w,h;

  w= bi_fdir_width;
  h= bi_fdir_height;

  fc= calloc(1,sizeof(*fc));
  fc->image= create_image(w,h);
  fc->w= w;
  fc->h= h;

  fc->func= finish;
  fc->arg= arg;

  attribs.background_pixmap= None;
  fc->window= XCreateWindow(x_display, parent, 0,0,w,h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);
  
  XSelectInput(x_display, fc->window, ExposureMask | 
                       ButtonPressMask | ButtonReleaseMask | 
                        StructureNotifyMask);
  register_event_handler(fc->window, Expose, fc, fc_expose, NULL);
  register_event_handler(fc->window, ButtonPress, fc, fc_press, NULL);
  register_event_handler(fc->window, ButtonRelease, fc, fc_release, NULL);
  register_event_handler(fc->window, DestroyNotify, fc, fc_destroy, NULL);

  fill_rect(fc->image->data, fc->w, fc->h, 0,0, fc->w, fc->h, 0xffffff);

  blend_alpha_image(fc->image->data, fc->w, fc->h,  0,0,
       (uint32_t*)  bi_fdir, bi_fdir_width, bi_fdir_height,
                    0,0, bi_fdir_width, bi_fdir_height, 0xffffff);
  
  fc->fs= file_selector_new(fc->window, 8,78, 621, 364,
                            dir, 0, NULL, click_file, fc);
  XMapWindow(x_display, fc->window);
}

