/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <stddef.h>
#include <stdio.h>

typedef struct {
  char *name;
  int kind;
  int offset, soffset, noffset, goffset;
  void *arg;
  int eltsize;
} cfgl_field_t;
typedef struct { char *name; int value; } cfgl_intcst_t;
typedef struct { char *name; char *value; } cfgl_strcst_t;
typedef struct { char *name; double value; } cfgl_flocst_t;


static char cfgl_ebuf[2000];
static jmp_buf cfgl_elabel;
static char *cfgl_fname;
static int spaces;



static int token;
static char *sbuf;
static int ssize= 100, slen;
static int ivalue;
static double fvalue;


static char* cfgl_parse_top(char *fn,void *ptr, cfgl_field_t *fields);
static void cfgl_parse_struct(void *ptr, cfgl_field_t *fields);
static void cfgl_parse_field(void *ptr, cfgl_field_t *fields);
static void cfgl_parse_int(int *ptr, char *given, cfgl_intcst_t *cst);
static void cfgl_lookup_int(cfgl_intcst_t *cst, int *ptr);
static void cfgl_lookup_flo(cfgl_flocst_t *cst, double *ptr);
static void cfgl_lookup_str(cfgl_strcst_t *cst, char **ptr);
static void cfgl_parse_intarray
   (int **ptr, int *size, int *len, cfgl_intcst_t *cst);
static void cfgl_put_int
  (int **ptr, int *size, int *len, cfgl_intcst_t *cst);
static void* cfgl_array_grow(void *ptr, int *size, int *len, int esize);
static void cfgl_parse_recarray
    (void **ptr, int *size, int *len, int esize, cfgl_field_t *fields);
static void cfgl_parse_flo(double *ptr, char *given, cfgl_flocst_t *cst);
static void cfgl_parse_floarray
   (double **ptr, int *size, int *len, cfgl_flocst_t *cst);
static void cfgl_put_flo
  (double **ptr, int *size, int *len, cfgl_flocst_t *cst);
static void cfgl_parse_str(char** ptr, char *given, cfgl_strcst_t *cst);
static void cfgl_parse_strarray
   (char ***ptr, int *size, int *len, cfgl_strcst_t *cst);
static void cfgl_put_str
  (char ***ptr, int *size, int *len, cfgl_strcst_t *cst);
static void cfgl_error(char *fmt, ...);
static void sreset() ;
static void sput(int x) ;
static char* cfgl_write_top(char *fn,void *ptr, cfgl_field_t *fields);
static void cfgl_write_field(FILE *out, void *ptr, cfgl_field_t *field);
static int has_valid_fields(void *ptr, cfgl_field_t* field);
static void cfgl_write_struct(FILE *f, void *ptr, cfgl_field_t *field);
static void cfgl_write_str(FILE *f, char *str);
static void print_spaces(FILE *out);
typedef struct cfg_page_prop {
  int window_border; char G_window_border;
  int window; char G_window;
  int root; char G_root;
} cfg_page_prop_t;
typedef struct cfg_pager_style {
  int border; char G_border;
  cfg_page_prop_t inactive;
  cfg_page_prop_t active;
} cfg_pager_style_t;
typedef struct cfg_clock {
  int font_size; char G_font_size;
  char* font_name; char G_font_name;
  int pad; char G_pad;
  int bg; char G_bg;
  int fg; char G_fg;
} cfg_clock_t;
typedef struct cfg_taskbar {
  int mid; char G_mid;
  int right; char G_right;
  int left; char G_left;
  int active; char G_active;
  int max_size; char G_max_size;
  int border; char G_border;
  int icon_size; char G_icon_size;
  int font_size; char G_font_size;
  char* font_name; char G_font_name;
  int bg; char G_bg;
  int fg; char G_fg;
} cfg_taskbar_t;
typedef struct cfg_menu_style {
  int hilight; char G_hilight;
  int sepcolor; char G_sepcolor;
  int sepv; char G_sepv;
  int seph; char G_seph;
  int fg; char G_fg;
  int bg; char G_bg;
  char* font_name; char G_font_name;
  int font_size; char G_font_size;
  int icon_h; char G_icon_h;
  int icon_w; char G_icon_w;
  int vspc; char G_vspc;
  int mid2; char G_mid2;
  int mid; char G_mid;
  int right; char G_right;
  int left; char G_left;
} cfg_menu_style_t;
typedef struct cfg_top {
  int override_redirect; char G_override_redirect;
  char* right_box_file; char G_right_box_file;
  char* left_box_file; char G_left_box_file;
  int button_size; char G_button_size;
  int spacing; char G_spacing;
  int left; char G_left;
  int right; char G_right;
  int fg; char G_fg;
  int bg; char G_bg;
  int height; char G_height;
  char* *menu_shortcut_path; int N_menu_shortcut_path; int S_menu_shortcut_path;
  char* *menu_icon_path; int N_menu_icon_path; int S_menu_icon_path;
  cfg_menu_style_t menu_style;
  int clock_on; char G_clock_on;
  cfg_clock_t clock;
  cfg_taskbar_t taskbar;
  cfg_pager_style_t pager_style;
} cfg_top_t;
static cfgl_intcst_t cfgl_global_intcst[]= {
 {0,0}
};
static cfgl_strcst_t cfgl_global_strcst[]= {
 {0,0}
};
static cfgl_flocst_t cfgl_global_flocst[]= {
 {0,0}
};
static cfgl_field_t cfgl_T1[]={
{"window_border", 'i' , offsetof(cfg_page_prop_t,window_border),  0,0,offsetof(cfg_page_prop_t,G_window_border),  0, 0},
{"window", 'i' , offsetof(cfg_page_prop_t,window),  0,0,offsetof(cfg_page_prop_t,G_window),  0, 0},
{"root", 'i' , offsetof(cfg_page_prop_t,root),  0,0,offsetof(cfg_page_prop_t,G_root),  0, 0},
 {0,0}
};
static cfgl_field_t cfgl_T2[]={
{"border", 'i' , offsetof(cfg_pager_style_t,border),  0,0,offsetof(cfg_pager_style_t,G_border),  0, 0},
{"inactive", 'r' , offsetof(cfg_pager_style_t,inactive),  0,0,0,  cfgl_T1, 0},
{"active", 'r' , offsetof(cfg_pager_style_t,active),  0,0,0,  cfgl_T1, 0},
 {0,0}
};
static cfgl_field_t cfgl_T3[]={
{"font_size", 'i' , offsetof(cfg_clock_t,font_size),  0,0,offsetof(cfg_clock_t,G_font_size),  0, 0},
{"font_name", 's' , offsetof(cfg_clock_t,font_name),  0,0,offsetof(cfg_clock_t,G_font_name),  0, 0},
{"pad", 'i' , offsetof(cfg_clock_t,pad),  0,0,offsetof(cfg_clock_t,G_pad),  0, 0},
{"bg", 'i' , offsetof(cfg_clock_t,bg),  0,0,offsetof(cfg_clock_t,G_bg),  0, 0},
{"fg", 'i' , offsetof(cfg_clock_t,fg),  0,0,offsetof(cfg_clock_t,G_fg),  0, 0},
 {0,0}
};
static cfgl_field_t cfgl_T4[]={
{"mid", 'i' , offsetof(cfg_taskbar_t,mid),  0,0,offsetof(cfg_taskbar_t,G_mid),  0, 0},
{"right", 'i' , offsetof(cfg_taskbar_t,right),  0,0,offsetof(cfg_taskbar_t,G_right),  0, 0},
{"left", 'i' , offsetof(cfg_taskbar_t,left),  0,0,offsetof(cfg_taskbar_t,G_left),  0, 0},
{"active", 'i' , offsetof(cfg_taskbar_t,active),  0,0,offsetof(cfg_taskbar_t,G_active),  0, 0},
{"max_size", 'i' , offsetof(cfg_taskbar_t,max_size),  0,0,offsetof(cfg_taskbar_t,G_max_size),  0, 0},
{"border", 'i' , offsetof(cfg_taskbar_t,border),  0,0,offsetof(cfg_taskbar_t,G_border),  0, 0},
{"icon_size", 'i' , offsetof(cfg_taskbar_t,icon_size),  0,0,offsetof(cfg_taskbar_t,G_icon_size),  0, 0},
{"font_size", 'i' , offsetof(cfg_taskbar_t,font_size),  0,0,offsetof(cfg_taskbar_t,G_font_size),  0, 0},
{"font_name", 's' , offsetof(cfg_taskbar_t,font_name),  0,0,offsetof(cfg_taskbar_t,G_font_name),  0, 0},
{"bg", 'i' , offsetof(cfg_taskbar_t,bg),  0,0,offsetof(cfg_taskbar_t,G_bg),  0, 0},
{"fg", 'i' , offsetof(cfg_taskbar_t,fg),  0,0,offsetof(cfg_taskbar_t,G_fg),  0, 0},
 {0,0}
};
static cfgl_field_t cfgl_T5[]={
{"hilight", 'i' , offsetof(cfg_menu_style_t,hilight),  0,0,offsetof(cfg_menu_style_t,G_hilight),  0, 0},
{"sepcolor", 'i' , offsetof(cfg_menu_style_t,sepcolor),  0,0,offsetof(cfg_menu_style_t,G_sepcolor),  0, 0},
{"sepv", 'i' , offsetof(cfg_menu_style_t,sepv),  0,0,offsetof(cfg_menu_style_t,G_sepv),  0, 0},
{"seph", 'i' , offsetof(cfg_menu_style_t,seph),  0,0,offsetof(cfg_menu_style_t,G_seph),  0, 0},
{"fg", 'i' , offsetof(cfg_menu_style_t,fg),  0,0,offsetof(cfg_menu_style_t,G_fg),  0, 0},
{"bg", 'i' , offsetof(cfg_menu_style_t,bg),  0,0,offsetof(cfg_menu_style_t,G_bg),  0, 0},
{"font_name", 's' , offsetof(cfg_menu_style_t,font_name),  0,0,offsetof(cfg_menu_style_t,G_font_name),  0, 0},
{"font_size", 'i' , offsetof(cfg_menu_style_t,font_size),  0,0,offsetof(cfg_menu_style_t,G_font_size),  0, 0},
{"icon_h", 'i' , offsetof(cfg_menu_style_t,icon_h),  0,0,offsetof(cfg_menu_style_t,G_icon_h),  0, 0},
{"icon_w", 'i' , offsetof(cfg_menu_style_t,icon_w),  0,0,offsetof(cfg_menu_style_t,G_icon_w),  0, 0},
{"vspc", 'i' , offsetof(cfg_menu_style_t,vspc),  0,0,offsetof(cfg_menu_style_t,G_vspc),  0, 0},
{"mid2", 'i' , offsetof(cfg_menu_style_t,mid2),  0,0,offsetof(cfg_menu_style_t,G_mid2),  0, 0},
{"mid", 'i' , offsetof(cfg_menu_style_t,mid),  0,0,offsetof(cfg_menu_style_t,G_mid),  0, 0},
{"right", 'i' , offsetof(cfg_menu_style_t,right),  0,0,offsetof(cfg_menu_style_t,G_right),  0, 0},
{"left", 'i' , offsetof(cfg_menu_style_t,left),  0,0,offsetof(cfg_menu_style_t,G_left),  0, 0},
 {0,0}
};
static cfgl_field_t cfgl_T6[]={
{"override_redirect", 'i' , offsetof(cfg_top_t,override_redirect),  0,0,offsetof(cfg_top_t,G_override_redirect),  0, 0},
{"right_box_file", 's' , offsetof(cfg_top_t,right_box_file),  0,0,offsetof(cfg_top_t,G_right_box_file),  0, 0},
{"left_box_file", 's' , offsetof(cfg_top_t,left_box_file),  0,0,offsetof(cfg_top_t,G_left_box_file),  0, 0},
{"button_size", 'i' , offsetof(cfg_top_t,button_size),  0,0,offsetof(cfg_top_t,G_button_size),  0, 0},
{"spacing", 'i' , offsetof(cfg_top_t,spacing),  0,0,offsetof(cfg_top_t,G_spacing),  0, 0},
{"left", 'i' , offsetof(cfg_top_t,left),  0,0,offsetof(cfg_top_t,G_left),  0, 0},
{"right", 'i' , offsetof(cfg_top_t,right),  0,0,offsetof(cfg_top_t,G_right),  0, 0},
{"fg", 'i' , offsetof(cfg_top_t,fg),  0,0,offsetof(cfg_top_t,G_fg),  0, 0},
{"bg", 'i' , offsetof(cfg_top_t,bg),  0,0,offsetof(cfg_top_t,G_bg),  0, 0},
{"height", 'i' , offsetof(cfg_top_t,height),  0,0,offsetof(cfg_top_t,G_height),  0, 0},
{"menu_shortcut_path", 'S' , offsetof(cfg_top_t,menu_shortcut_path), offsetof(cfg_top_t,S_menu_shortcut_path), offsetof(cfg_top_t,N_menu_shortcut_path),0,  0, 0},
{"menu_icon_path", 'S' , offsetof(cfg_top_t,menu_icon_path), offsetof(cfg_top_t,S_menu_icon_path), offsetof(cfg_top_t,N_menu_icon_path),0,  0, 0},
{"menu_style", 'r' , offsetof(cfg_top_t,menu_style),  0,0,0,  cfgl_T5, 0},
{"clock_on", 'i' , offsetof(cfg_top_t,clock_on),  0,0,offsetof(cfg_top_t,G_clock_on),  0, 0},
{"clock", 'r' , offsetof(cfg_top_t,clock),  0,0,0,  cfgl_T3, 0},
{"taskbar", 'r' , offsetof(cfg_top_t,taskbar),  0,0,0,  cfgl_T4, 0},
{"pager_style", 'r' , offsetof(cfg_top_t,pager_style),  0,0,0,  cfgl_T2, 0},
 {0,0}
};
/*** THIS IS THE FLEX START ***/


#define  YY_INT_ALIGNED short int

/* A lexical scanner generated by flex */

#define yy_create_buffer cfgl__create_buffer
#define yy_delete_buffer cfgl__delete_buffer
#define yy_flex_debug cfgl__flex_debug
#define yy_init_buffer cfgl__init_buffer
#define yy_flush_buffer cfgl__flush_buffer
#define yy_load_buffer_state cfgl__load_buffer_state
#define yy_switch_to_buffer cfgl__switch_to_buffer
#define yyin cfgl_in
#define yyleng cfgl_leng
#define yylex cfgl_lex
#define yylineno cfgl_lineno
#define yyout cfgl_out
#define yyrestart cfgl_restart
#define yytext cfgl_text
#define yywrap cfgl_wrap
#define yyalloc cfgl_alloc
#define yyrealloc cfgl_realloc
#define yyfree cfgl_free

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 5
#define YY_FLEX_SUBMINOR_VERSION 37
#if YY_FLEX_SUBMINOR_VERSION > 0
#define FLEX_BETA
#endif

/* First, we deal with  platform-specific or compiler-specific issues. */

/* begin standard C headers. */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

/* end standard C headers. */

/* flex integer type definitions */

#ifndef FLEXINT_H
#define FLEXINT_H

/* C99 systems have <inttypes.h>. Non-C99 systems may or may not. */

#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L

/* C99 says to define __STDC_LIMIT_MACROS before including stdint.h,
 * if you want the limit (max/min) macros for int types. 
 */
#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS 1
#endif

#include <inttypes.h>
typedef int8_t flex_int8_t;
typedef uint8_t flex_uint8_t;
typedef int16_t flex_int16_t;
typedef uint16_t flex_uint16_t;
typedef int32_t flex_int32_t;
typedef uint32_t flex_uint32_t;
#else
typedef signed char flex_int8_t;
typedef short int flex_int16_t;
typedef int flex_int32_t;
typedef unsigned char flex_uint8_t; 
typedef unsigned short int flex_uint16_t;
typedef unsigned int flex_uint32_t;

/* Limits of integral types. */
#ifndef INT8_MIN
#define INT8_MIN               (-128)
#endif
#ifndef INT16_MIN
#define INT16_MIN              (-32767-1)
#endif
#ifndef INT32_MIN
#define INT32_MIN              (-2147483647-1)
#endif
#ifndef INT8_MAX
#define INT8_MAX               (127)
#endif
#ifndef INT16_MAX
#define INT16_MAX              (32767)
#endif
#ifndef INT32_MAX
#define INT32_MAX              (2147483647)
#endif
#ifndef UINT8_MAX
#define UINT8_MAX              (255U)
#endif
#ifndef UINT16_MAX
#define UINT16_MAX             (65535U)
#endif
#ifndef UINT32_MAX
#define UINT32_MAX             (4294967295U)
#endif

#endif /* ! C99 */

#endif /* ! FLEXINT_H */

#ifdef __cplusplus

/* The "const" storage-class-modifier is valid. */
#define YY_USE_CONST

#else	/* ! __cplusplus */

/* C99 requires __STDC__ to be defined as 1. */
#if defined (__STDC__)

#define YY_USE_CONST

#endif	/* defined (__STDC__) */
#endif	/* ! __cplusplus */

#ifdef YY_USE_CONST
#define yyconst const
#else
#define yyconst
#endif

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an unsigned
 * integer for use as an array index.  If the signed char is negative,
 * we want to instead treat it as an 8-bit unsigned char, hence the
 * double cast.
 */
#define YY_SC_TO_UI(c) ((unsigned int) (unsigned char) c)

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN (yy_start) = 1 + 2 *

/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START (((yy_start) - 1) / 2)
#define YYSTATE YY_START

/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)

/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE cfgl_restart(cfgl_in  )

#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#ifndef YY_BUF_SIZE
#define YY_BUF_SIZE 16384
#endif

/* The state buf must be large enough to hold one state per character in the main buffer.
 */
#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(yy_state_type))

#ifndef YY_TYPEDEF_YY_BUFFER_STATE
#define YY_TYPEDEF_YY_BUFFER_STATE
typedef struct yy_buffer_state *YY_BUFFER_STATE;
#endif

#ifndef YY_TYPEDEF_YY_SIZE_T
#define YY_TYPEDEF_YY_SIZE_T
typedef size_t yy_size_t;
#endif

extern yy_size_t cfgl_leng;

extern FILE *cfgl_in, *cfgl_out;

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2

    /* Note: We specifically omit the test for yy_rule_can_match_eol because it requires
     *       access to the local variable yy_act. Since yyless() is a macro, it would break
     *       existing scanners that call yyless() from OUTSIDE cfgl_lex. 
     *       One obvious solution it to make yy_act a global. I tried that, and saw
     *       a 5% performance hit in a non-cfgl_lineno scanner, because yy_act is
     *       normally declared as a register variable-- so it is not worth it.
     */
    #define  YY_LESS_LINENO(n) \
            do { \
                int yyl;\
                for ( yyl = n; yyl < cfgl_leng; ++yyl )\
                    if ( cfgl_text[yyl] == '\n' )\
                        --cfgl_lineno;\
            }while(0)
    
/* Return all but the first "n" matched characters back to the input stream. */
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up cfgl_text. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		*yy_cp = (yy_hold_char); \
		YY_RESTORE_YY_MORE_OFFSET \
		(yy_c_buf_p) = yy_cp = yy_bp + yyless_macro_arg - YY_MORE_ADJ; \
		YY_DO_BEFORE_ACTION; /* set up cfgl_text again */ \
		} \
	while ( 0 )

#define unput(c) yyunput( c, (yytext_ptr)  )

#ifndef YY_STRUCT_YY_BUFFER_STATE
#define YY_STRUCT_YY_BUFFER_STATE
struct yy_buffer_state
	{
	FILE *yy_input_file;

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	yy_size_t yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	yy_size_t yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

    int yy_bs_lineno; /**< The line count. */
    int yy_bs_column; /**< The column count. */
    
	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via cfgl_restart()), so that the user can continue scanning by
	 * just pointing cfgl_in at a new input file.
	 */
#define YY_BUFFER_EOF_PENDING 2

	};
#endif /* !YY_STRUCT_YY_BUFFER_STATE */

/* Stack of input buffers. */
static size_t yy_buffer_stack_top = 0; /**< index of top of stack. */
static size_t yy_buffer_stack_max = 0; /**< capacity of stack. */
static YY_BUFFER_STATE * yy_buffer_stack = 0; /**< Stack as an array. */

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 *
 * Returns the top of the stack, or NULL.
 */
#define YY_CURRENT_BUFFER ( (yy_buffer_stack) \
                          ? (yy_buffer_stack)[(yy_buffer_stack_top)] \
                          : NULL)

/* Same as previous macro, but useful when we know that the buffer stack is not
 * NULL or when we need an lvalue. For internal use only.
 */
#define YY_CURRENT_BUFFER_LVALUE (yy_buffer_stack)[(yy_buffer_stack_top)]

/* yy_hold_char holds the character lost when cfgl_text is formed. */
static char yy_hold_char;
static yy_size_t yy_n_chars;		/* number of characters read into yy_ch_buf */
yy_size_t cfgl_leng;

/* Points to current character in buffer. */
static char *yy_c_buf_p = (char *) 0;
static int yy_init = 0;		/* whether we need to initialize */
static int yy_start = 0;	/* start state number */

/* Flag which is used to allow cfgl_wrap()'s to do buffer switches
 * instead of setting up a fresh cfgl_in.  A bit of a hack ...
 */
static int yy_did_buffer_switch_on_eof;

void cfgl_restart (FILE *input_file  );
void cfgl__switch_to_buffer (YY_BUFFER_STATE new_buffer  );
YY_BUFFER_STATE cfgl__create_buffer (FILE *file,int size  );
void cfgl__delete_buffer (YY_BUFFER_STATE b  );
void cfgl__flush_buffer (YY_BUFFER_STATE b  );
void cfgl_push_buffer_state (YY_BUFFER_STATE new_buffer  );
void cfgl_pop_buffer_state (void );

static void cfgl_ensure_buffer_stack (void );
static void cfgl__load_buffer_state (void );
static void cfgl__init_buffer (YY_BUFFER_STATE b,FILE *file  );

#define YY_FLUSH_BUFFER cfgl__flush_buffer(YY_CURRENT_BUFFER )

YY_BUFFER_STATE cfgl__scan_buffer (char *base,yy_size_t size  );
YY_BUFFER_STATE cfgl__scan_string (yyconst char *yy_str  );
YY_BUFFER_STATE cfgl__scan_bytes (yyconst char *bytes,yy_size_t len  );

void *cfgl_alloc (yy_size_t  );
void *cfgl_realloc (void *,yy_size_t  );
void cfgl_free (void *  );

#define yy_new_buffer cfgl__create_buffer

#define yy_set_interactive(is_interactive) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){ \
        cfgl_ensure_buffer_stack (); \
		YY_CURRENT_BUFFER_LVALUE =    \
            cfgl__create_buffer(cfgl_in,YY_BUF_SIZE ); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_is_interactive = is_interactive; \
	}

#define yy_set_bol(at_bol) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){\
        cfgl_ensure_buffer_stack (); \
		YY_CURRENT_BUFFER_LVALUE =    \
            cfgl__create_buffer(cfgl_in,YY_BUF_SIZE ); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_at_bol = at_bol; \
	}

#define YY_AT_BOL() (YY_CURRENT_BUFFER_LVALUE->yy_at_bol)

/* Begin user sect3 */

#define cfgl_wrap() 1
#define YY_SKIP_YYWRAP

typedef unsigned char YY_CHAR;

FILE *cfgl_in = (FILE *) 0, *cfgl_out = (FILE *) 0;

typedef int yy_state_type;

extern int cfgl_lineno;

int cfgl_lineno = 1;

extern char *cfgl_text;
#define yytext_ptr cfgl_text

static yy_state_type yy_get_previous_state (void );
static yy_state_type yy_try_NUL_trans (yy_state_type current_state  );
static int yy_get_next_buffer (void );
static void yy_fatal_error (yyconst char msg[]  );

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up cfgl_text.
 */
#define YY_DO_BEFORE_ACTION \
	(yytext_ptr) = yy_bp; \
	cfgl_leng = (size_t) (yy_cp - yy_bp); \
	(yy_hold_char) = *yy_cp; \
	*yy_cp = '\0'; \
	(yy_c_buf_p) = yy_cp;

#define YY_NUM_RULES 24
#define YY_END_OF_BUFFER 25
/* This struct is not used in this scanner,
   but its presence is necessary. */
struct yy_trans_info
	{
	flex_int32_t yy_verify;
	flex_int32_t yy_nxt;
	};
static yyconst flex_int16_t yy_accept[50] =
    {   0,
        0,    0,    0,    0,    0,    0,   25,    1,    1,   23,
        2,    4,    3,   23,    9,    9,    7,    5,    6,   22,
       24,   11,   22,   12,    1,    4,    9,    0,    9,    0,
        7,   21,   13,   14,   15,   16,   17,   18,   19,   21,
        9,    8,   10,    0,    0,   20,    0,    8,    0
    } ;

static yyconst flex_int32_t yy_ec[256] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    2,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    3,    4,    5,    3,    3,    3,    6,    3,
        3,    3,    7,    3,    8,    9,    3,   10,   11,   11,
       11,   11,   11,   11,   11,   11,   11,    3,    3,    3,
        3,    3,    3,    3,   12,   12,   12,   12,   13,   12,
       14,   14,   14,   14,   14,   14,   14,   14,   14,   14,
       14,   14,   14,   14,   14,   14,   14,   14,   14,   14,
        3,   15,    3,    3,   14,    3,   16,   17,   12,   12,

       13,   18,   14,   14,   14,   14,   14,   14,   14,   19,
       14,   14,   14,   20,   14,   21,   14,   22,   14,   23,
       14,   14,   24,    3,   25,    3,    1,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,

        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3
    } ;

static yyconst flex_int32_t yy_meta[26] =
    {   0,
        1,    2,    1,    1,    1,    1,    1,    1,    1,    3,
        3,    3,    3,    4,    1,    3,    3,    3,    4,    4,
        4,    4,    4,    1,    1
    } ;

static yyconst flex_int16_t yy_base[57] =
    {   0,
        0,    0,   24,   25,   28,   29,   95,   31,   35,   96,
       96,    0,   96,   31,   36,   39,    0,   96,   96,   96,
       96,   96,   44,   96,   50,    0,   43,   45,   59,    0,
        0,   96,   96,   96,   96,   96,   96,   96,   96,    0,
       47,   61,    0,    0,   68,   96,   67,   69,   96,   80,
       84,   86,   90,   83,   70,   35
    } ;

static yyconst flex_int16_t yy_def[57] =
    {   0,
       49,    1,   50,   50,   50,   50,   49,   49,   49,   49,
       49,   51,   49,   49,   49,   49,   52,   49,   49,   49,
       49,   49,   53,   49,   49,   51,   49,   49,   49,   54,
       52,   49,   49,   49,   49,   49,   49,   49,   49,   55,
       49,   49,   54,   56,   49,   49,   49,   49,    0,   49,
       49,   49,   49,   49,   49,   49
    } ;

static yyconst flex_int16_t yy_nxt[122] =
    {   0,
        8,    9,   10,   11,   12,   13,   10,   14,   10,   15,
       16,   17,   17,   17,   10,   17,   17,   17,   17,   17,
       17,   17,   17,   18,   19,   21,   21,   22,   22,   21,
       21,   25,   25,   24,   24,   25,   25,   46,   23,   23,
       27,   27,   23,   23,   28,   29,   29,   28,   29,   29,
       25,   25,   41,   41,   42,   42,   41,   41,   30,   33,
       34,   35,   36,   37,   38,   39,   40,   28,   29,   29,
       42,   42,   44,   45,   47,   47,   48,   48,   48,   48,
       20,   20,   20,   20,   26,   43,   26,   26,   31,   31,
       32,   32,   32,   32,   49,    7,   49,   49,   49,   49,

       49,   49,   49,   49,   49,   49,   49,   49,   49,   49,
       49,   49,   49,   49,   49,   49,   49,   49,   49,   49,
       49
    } ;

static yyconst flex_int16_t yy_chk[122] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    3,    4,    3,    4,    5,
        6,    8,    8,    5,    6,    9,    9,   56,    3,    4,
       14,   14,    5,    6,   15,   15,   15,   16,   16,   16,
       25,   25,   27,   27,   28,   28,   41,   41,   15,   23,
       23,   23,   23,   23,   23,   23,   23,   29,   29,   29,
       42,   42,   55,   42,   45,   45,   47,   47,   48,   48,
       50,   50,   50,   50,   51,   54,   51,   51,   52,   52,
       53,   53,   53,   53,    7,   49,   49,   49,   49,   49,

       49,   49,   49,   49,   49,   49,   49,   49,   49,   49,
       49,   49,   49,   49,   49,   49,   49,   49,   49,   49,
       49
    } ;

/* Table of booleans, true if rule could match eol. */
static yyconst flex_int32_t yy_rule_can_match_eol[25] =
    {   0,
1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 1, 0, 0, 0,     };

static yy_state_type yy_last_accepting_state;
static char *yy_last_accepting_cpos;

extern int cfgl__flex_debug;
int cfgl__flex_debug = 0;

/* The intent behind this definition is that it'll catch
 * any uses of REJECT which flex missed.
 */
#define REJECT reject_used_but_not_detected
#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
char *cfgl_text;
#define YY_NO_INPUT 1

#define INITIAL 0
#define dquote 1
#define squote 2

#ifndef YY_NO_UNISTD_H
/* Special case for "unistd.h", since it is non-ANSI. We include it way
 * down here because we want the user's section 1 to have been scanned first.
 * The user has a chance to override it with an option.
 */
#include <unistd.h>
#endif

#ifndef YY_EXTRA_TYPE
#define YY_EXTRA_TYPE void *
#endif

static int yy_init_globals (void );

/* Accessor methods to globals.
   These are made visible to non-reentrant scanners for convenience. */

int cfgl_lex_destroy (void );

int cfgl_get_debug (void );

void cfgl_set_debug (int debug_flag  );

YY_EXTRA_TYPE cfgl_get_extra (void );

void cfgl_set_extra (YY_EXTRA_TYPE user_defined  );

FILE *cfgl_get_in (void );

void cfgl_set_in  (FILE * in_str  );

FILE *cfgl_get_out (void );

void cfgl_set_out  (FILE * out_str  );

yy_size_t cfgl_get_leng (void );

char *cfgl_get_text (void );

int cfgl_get_lineno (void );

void cfgl_set_lineno (int line_number  );

/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int cfgl_wrap (void );
#else
extern int cfgl_wrap (void );
#endif
#endif

#ifndef yytext_ptr
static void yy_flex_strncpy (char *,yyconst char *,int );
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (yyconst char * );
#endif

#ifndef YY_NO_INPUT

#ifdef __cplusplus
static int yyinput (void );
#else
static int input (void );
#endif

#endif

/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#define YY_READ_BUF_SIZE 8192
#endif

/* Copy whatever the last rule matched to the standard output. */
#ifndef ECHO
/* This used to be an fputs(), but since the string might contain NUL's,
 * we now use fwrite().
 */
#define ECHO do { if (fwrite( cfgl_text, cfgl_leng, 1, cfgl_out )) {} } while (0)
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if ( YY_CURRENT_BUFFER_LVALUE->yy_is_interactive ) \
		{ \
		int c = '*'; \
		size_t n; \
		for ( n = 0; n < max_size && \
			     (c = getc( cfgl_in )) != EOF && c != '\n'; ++n ) \
			buf[n] = (char) c; \
		if ( c == '\n' ) \
			buf[n++] = (char) c; \
		if ( c == EOF && ferror( cfgl_in ) ) \
			YY_FATAL_ERROR( "input in flex scanner failed" ); \
		result = n; \
		} \
	else \
		{ \
		errno=0; \
		while ( (result = fread(buf, 1, max_size, cfgl_in))==0 && ferror(cfgl_in)) \
			{ \
			if( errno != EINTR) \
				{ \
				YY_FATAL_ERROR( "input in flex scanner failed" ); \
				break; \
				} \
			errno=0; \
			clearerr(cfgl_in); \
			} \
		}\
\

#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) yy_fatal_error( msg )
#endif

/* end tables serialization structures and prototypes */

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL_IS_OURS 1

extern int cfgl_lex (void);

#define YY_DECL int cfgl_lex (void)
#endif /* !YY_DECL */

/* Code executed at the beginning of each rule, after cfgl_text and cfgl_leng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK break;
#endif

#define YY_RULE_SETUP \
	YY_USER_ACTION

/** The main scanner function which does all the work.
 */
YY_DECL
{
	register yy_state_type yy_current_state;
	register char *yy_cp, *yy_bp;
	register int yy_act;
    
	if ( !(yy_init) )
		{
		(yy_init) = 1;

#ifdef YY_USER_INIT
		YY_USER_INIT;
#endif

		if ( ! (yy_start) )
			(yy_start) = 1;	/* first start state */

		if ( ! cfgl_in )
			cfgl_in = stdin;

		if ( ! cfgl_out )
			cfgl_out = stdout;

		if ( ! YY_CURRENT_BUFFER ) {
			cfgl_ensure_buffer_stack ();
			YY_CURRENT_BUFFER_LVALUE =
				cfgl__create_buffer(cfgl_in,YY_BUF_SIZE );
		}

		cfgl__load_buffer_state( );
		}

	while ( 1 )		/* loops until end-of-file is reached */
		{
		yy_cp = (yy_c_buf_p);

		/* Support of cfgl_text. */
		*yy_cp = (yy_hold_char);

		/* yy_bp points to the position in yy_ch_buf of the start of
		 * the current run.
		 */
		yy_bp = yy_cp;

		yy_current_state = (yy_start);
yy_match:
		do
			{
			register YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)];
			if ( yy_accept[yy_current_state] )
				{
				(yy_last_accepting_state) = yy_current_state;
				(yy_last_accepting_cpos) = yy_cp;
				}
			while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
				{
				yy_current_state = (int) yy_def[yy_current_state];
				if ( yy_current_state >= 50 )
					yy_c = yy_meta[(unsigned int) yy_c];
				}
			yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
			++yy_cp;
			}
		while ( yy_base[yy_current_state] != 96 );

yy_find_action:
		yy_act = yy_accept[yy_current_state];
		if ( yy_act == 0 )
			{ /* have to back up */
			yy_cp = (yy_last_accepting_cpos);
			yy_current_state = (yy_last_accepting_state);
			yy_act = yy_accept[yy_current_state];
			}

		YY_DO_BEFORE_ACTION;

		if ( yy_act != YY_END_OF_BUFFER && yy_rule_can_match_eol[yy_act] )
			{
			int yyl;
			for ( yyl = 0; yyl < cfgl_leng; ++yyl )
				if ( cfgl_text[yyl] == '\n' )
					   
    cfgl_lineno++;
;
			}

do_action:	/* This label is used only to access EOF actions. */

		switch ( yy_act )
	{ /* beginning of action switch */
			case 0: /* must back up */
			/* undo the effects of YY_DO_BEFORE_ACTION */
			*yy_cp = (yy_hold_char);
			yy_cp = (yy_last_accepting_cpos);
			yy_current_state = (yy_last_accepting_state);
			goto yy_find_action;

case 1:
/* rule 1 can match eol */
YY_RULE_SETUP
{ }
	YY_BREAK
case 2:
YY_RULE_SETUP
{ BEGIN(dquote); sreset(); }
	YY_BREAK
case 3:
YY_RULE_SETUP
{ BEGIN(squote); sreset(); }
	YY_BREAK
case 4:
YY_RULE_SETUP
{ /* commentos */ }
	YY_BREAK
case 5:
YY_RULE_SETUP
{ token= '{'; return 0; }
	YY_BREAK
case 6:
YY_RULE_SETUP
{ token= '}'; return 0; }
	YY_BREAK
case 7:
YY_RULE_SETUP
{ token= 'A'; return 0; }
	YY_BREAK
case 8:
YY_RULE_SETUP
{ fvalue= strtod(cfgl_text,0);
                                   token= 'F'; return 0; }
	YY_BREAK
case 9:
YY_RULE_SETUP
{ ivalue= strtol(cfgl_text,0,10); token= 'I'; return 0;}
	YY_BREAK
case 10:
YY_RULE_SETUP
{ ivalue= strtol(cfgl_text,0,0); token= 'I'; return 0; }
	YY_BREAK
case 11:
YY_RULE_SETUP
{ BEGIN(0); sbuf[slen]= 0; token= 'S'; return 0; }
	YY_BREAK
case 12:
YY_RULE_SETUP
{ BEGIN(0); sbuf[slen]= 0; token= 'S'; return 0;}
	YY_BREAK
case 13:
YY_RULE_SETUP
{ sput('\a'); }
	YY_BREAK
case 14:
YY_RULE_SETUP
{ sput('\b'); }
	YY_BREAK
case 15:
YY_RULE_SETUP
{ sput('\f'); }
	YY_BREAK
case 16:
YY_RULE_SETUP
{ sput('\n'); }
	YY_BREAK
case 17:
YY_RULE_SETUP
{ sput('\r'); }
	YY_BREAK
case 18:
YY_RULE_SETUP
{ sput('\t'); }
	YY_BREAK
case 19:
YY_RULE_SETUP
{ sput('\v'); }
	YY_BREAK
case 20:
YY_RULE_SETUP
{
           char buf[7];
           sprintf(buf, "0%s", cfgl_text+1);
           sput( strtol(buf,0,0)); };
	YY_BREAK
case 21:
/* rule 21 can match eol */
YY_RULE_SETUP
{ sput(cfgl_text[1]); }
	YY_BREAK
case 22:
YY_RULE_SETUP
{ sput(cfgl_text[0]); }
	YY_BREAK
case 23:
YY_RULE_SETUP
{ cfgl_error("unhandled char (%d)", cfgl_text[0]); }
	YY_BREAK
case YY_STATE_EOF(INITIAL):
case YY_STATE_EOF(dquote):
case YY_STATE_EOF(squote):
{ token= 'E'; return 0; }
	YY_BREAK
case 24:
YY_RULE_SETUP
ECHO;
	YY_BREAK

	case YY_END_OF_BUFFER:
		{
		/* Amount of text matched not including the EOB char. */
		int yy_amount_of_matched_text = (int) (yy_cp - (yytext_ptr)) - 1;

		/* Undo the effects of YY_DO_BEFORE_ACTION. */
		*yy_cp = (yy_hold_char);
		YY_RESTORE_YY_MORE_OFFSET

		if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_NEW )
			{
			/* We're scanning a new file or input source.  It's
			 * possible that this happened because the user
			 * just pointed cfgl_in at a new source and called
			 * cfgl_lex().  If so, then we have to assure
			 * consistency between YY_CURRENT_BUFFER and our
			 * globals.  Here is the right place to do so, because
			 * this is the first action (other than possibly a
			 * back-up) that will match for the new input source.
			 */
			(yy_n_chars) = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
			YY_CURRENT_BUFFER_LVALUE->yy_input_file = cfgl_in;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_NORMAL;
			}

		/* Note that here we test for yy_c_buf_p "<=" to the position
		 * of the first EOB in the buffer, since yy_c_buf_p will
		 * already have been incremented past the NUL character
		 * (since all states make transitions on EOB to the
		 * end-of-buffer state).  Contrast this with the test
		 * in input().
		 */
		if ( (yy_c_buf_p) <= &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)] )
			{ /* This was really a NUL. */
			yy_state_type yy_next_state;

			(yy_c_buf_p) = (yytext_ptr) + yy_amount_of_matched_text;

			yy_current_state = yy_get_previous_state(  );

			/* Okay, we're now positioned to make the NUL
			 * transition.  We couldn't have
			 * yy_get_previous_state() go ahead and do it
			 * for us because it doesn't know how to deal
			 * with the possibility of jamming (and we don't
			 * want to build jamming into it because then it
			 * will run more slowly).
			 */

			yy_next_state = yy_try_NUL_trans( yy_current_state );

			yy_bp = (yytext_ptr) + YY_MORE_ADJ;

			if ( yy_next_state )
				{
				/* Consume the NUL. */
				yy_cp = ++(yy_c_buf_p);
				yy_current_state = yy_next_state;
				goto yy_match;
				}

			else
				{
				yy_cp = (yy_c_buf_p);
				goto yy_find_action;
				}
			}

		else switch ( yy_get_next_buffer(  ) )
			{
			case EOB_ACT_END_OF_FILE:
				{
				(yy_did_buffer_switch_on_eof) = 0;

				if ( cfgl_wrap( ) )
					{
					/* Note: because we've taken care in
					 * yy_get_next_buffer() to have set up
					 * cfgl_text, we can now set up
					 * yy_c_buf_p so that if some total
					 * hoser (like flex itself) wants to
					 * call the scanner after we return the
					 * YY_NULL, it'll still work - another
					 * YY_NULL will get returned.
					 */
					(yy_c_buf_p) = (yytext_ptr) + YY_MORE_ADJ;

					yy_act = YY_STATE_EOF(YY_START);
					goto do_action;
					}

				else
					{
					if ( ! (yy_did_buffer_switch_on_eof) )
						YY_NEW_FILE;
					}
				break;
				}

			case EOB_ACT_CONTINUE_SCAN:
				(yy_c_buf_p) =
					(yytext_ptr) + yy_amount_of_matched_text;

				yy_current_state = yy_get_previous_state(  );

				yy_cp = (yy_c_buf_p);
				yy_bp = (yytext_ptr) + YY_MORE_ADJ;
				goto yy_match;

			case EOB_ACT_LAST_MATCH:
				(yy_c_buf_p) =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)];

				yy_current_state = yy_get_previous_state(  );

				yy_cp = (yy_c_buf_p);
				yy_bp = (yytext_ptr) + YY_MORE_ADJ;
				goto yy_find_action;
			}
		break;
		}

	default:
		YY_FATAL_ERROR(
			"fatal flex scanner internal error--no action found" );
	} /* end of action switch */
		} /* end of scanning one token */
} /* end of cfgl_lex */

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */
static int yy_get_next_buffer (void)
{
    	register char *dest = YY_CURRENT_BUFFER_LVALUE->yy_ch_buf;
	register char *source = (yytext_ptr);
	register int number_to_move, i;
	int ret_val;

	if ( (yy_c_buf_p) > &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars) + 1] )
		YY_FATAL_ERROR(
		"fatal flex scanner internal error--end of buffer missed" );

	if ( YY_CURRENT_BUFFER_LVALUE->yy_fill_buffer == 0 )
		{ /* Don't try to fill the buffer, so this is an EOF. */
		if ( (yy_c_buf_p) - (yytext_ptr) - YY_MORE_ADJ == 1 )
			{
			/* We matched a single character, the EOB, so
			 * treat this as a final EOF.
			 */
			return EOB_ACT_END_OF_FILE;
			}

		else
			{
			/* We matched some text prior to the EOB, first
			 * process it.
			 */
			return EOB_ACT_LAST_MATCH;
			}
		}

	/* Try to read more data. */

	/* First move last chars to start of buffer. */
	number_to_move = (int) ((yy_c_buf_p) - (yytext_ptr)) - 1;

	for ( i = 0; i < number_to_move; ++i )
		*(dest++) = *(source++);

	if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_EOF_PENDING )
		/* don't do the read, it's not guaranteed to return an EOF,
		 * just force an EOF
		 */
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars) = 0;

	else
		{
			yy_size_t num_to_read =
			YY_CURRENT_BUFFER_LVALUE->yy_buf_size - number_to_move - 1;

		while ( num_to_read <= 0 )
			{ /* Not enough room in the buffer - grow it. */

			/* just a shorter name for the current buffer */
			YY_BUFFER_STATE b = YY_CURRENT_BUFFER_LVALUE;

			int yy_c_buf_p_offset =
				(int) ((yy_c_buf_p) - b->yy_ch_buf);

			if ( b->yy_is_our_buffer )
				{
				yy_size_t new_size = b->yy_buf_size * 2;

				if ( new_size <= 0 )
					b->yy_buf_size += b->yy_buf_size / 8;
				else
					b->yy_buf_size *= 2;

				b->yy_ch_buf = (char *)
					/* Include room in for 2 EOB chars. */
					cfgl_realloc((void *) b->yy_ch_buf,b->yy_buf_size + 2  );
				}
			else
				/* Can't grow it, we don't own it. */
				b->yy_ch_buf = 0;

			if ( ! b->yy_ch_buf )
				YY_FATAL_ERROR(
				"fatal error - scanner input buffer overflow" );

			(yy_c_buf_p) = &b->yy_ch_buf[yy_c_buf_p_offset];

			num_to_read = YY_CURRENT_BUFFER_LVALUE->yy_buf_size -
						number_to_move - 1;

			}

		if ( num_to_read > YY_READ_BUF_SIZE )
			num_to_read = YY_READ_BUF_SIZE;

		/* Read in more data. */
		YY_INPUT( (&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move]),
			(yy_n_chars), num_to_read );

		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars);
		}

	if ( (yy_n_chars) == 0 )
		{
		if ( number_to_move == YY_MORE_ADJ )
			{
			ret_val = EOB_ACT_END_OF_FILE;
			cfgl_restart(cfgl_in  );
			}

		else
			{
			ret_val = EOB_ACT_LAST_MATCH;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status =
				YY_BUFFER_EOF_PENDING;
			}
		}

	else
		ret_val = EOB_ACT_CONTINUE_SCAN;

	if ((yy_size_t) ((yy_n_chars) + number_to_move) > YY_CURRENT_BUFFER_LVALUE->yy_buf_size) {
		/* Extend the array by 50%, plus the number we really need. */
		yy_size_t new_size = (yy_n_chars) + number_to_move + ((yy_n_chars) >> 1);
		YY_CURRENT_BUFFER_LVALUE->yy_ch_buf = (char *) cfgl_realloc((void *) YY_CURRENT_BUFFER_LVALUE->yy_ch_buf,new_size  );
		if ( ! YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			YY_FATAL_ERROR( "out of dynamic memory in yy_get_next_buffer()" );
	}

	(yy_n_chars) += number_to_move;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)] = YY_END_OF_BUFFER_CHAR;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars) + 1] = YY_END_OF_BUFFER_CHAR;

	(yytext_ptr) = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[0];

	return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

    static yy_state_type yy_get_previous_state (void)
{
	register yy_state_type yy_current_state;
	register char *yy_cp;
    
	yy_current_state = (yy_start);

	for ( yy_cp = (yytext_ptr) + YY_MORE_ADJ; yy_cp < (yy_c_buf_p); ++yy_cp )
		{
		register YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 3);
		if ( yy_accept[yy_current_state] )
			{
			(yy_last_accepting_state) = yy_current_state;
			(yy_last_accepting_cpos) = yy_cp;
			}
		while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
			{
			yy_current_state = (int) yy_def[yy_current_state];
			if ( yy_current_state >= 50 )
				yy_c = yy_meta[(unsigned int) yy_c];
			}
		yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
		}

	return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */
    static yy_state_type yy_try_NUL_trans  (yy_state_type yy_current_state )
{
	register int yy_is_jam;
    	register char *yy_cp = (yy_c_buf_p);

	register YY_CHAR yy_c = 3;
	if ( yy_accept[yy_current_state] )
		{
		(yy_last_accepting_state) = yy_current_state;
		(yy_last_accepting_cpos) = yy_cp;
		}
	while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		yy_current_state = (int) yy_def[yy_current_state];
		if ( yy_current_state >= 50 )
			yy_c = yy_meta[(unsigned int) yy_c];
		}
	yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
	yy_is_jam = (yy_current_state == 49);

		return yy_is_jam ? 0 : yy_current_state;
}

#ifndef YY_NO_INPUT
#ifdef __cplusplus
    static int yyinput (void)
#else
    static int input  (void)
#endif

{
	int c;
    
	*(yy_c_buf_p) = (yy_hold_char);

	if ( *(yy_c_buf_p) == YY_END_OF_BUFFER_CHAR )
		{
		/* yy_c_buf_p now points to the character we want to return.
		 * If this occurs *before* the EOB characters, then it's a
		 * valid NUL; if not, then we've hit the end of the buffer.
		 */
		if ( (yy_c_buf_p) < &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)] )
			/* This was really a NUL. */
			*(yy_c_buf_p) = '\0';

		else
			{ /* need more input */
			yy_size_t offset = (yy_c_buf_p) - (yytext_ptr);
			++(yy_c_buf_p);

			switch ( yy_get_next_buffer(  ) )
				{
				case EOB_ACT_LAST_MATCH:
					/* This happens because yy_g_n_b()
					 * sees that we've accumulated a
					 * token and flags that we need to
					 * try matching the token before
					 * proceeding.  But for input(),
					 * there's no matching to consider.
					 * So convert the EOB_ACT_LAST_MATCH
					 * to EOB_ACT_END_OF_FILE.
					 */

					/* Reset buffer status. */
					cfgl_restart(cfgl_in );

					/*FALLTHROUGH*/

				case EOB_ACT_END_OF_FILE:
					{
					if ( cfgl_wrap( ) )
						return EOF;

					if ( ! (yy_did_buffer_switch_on_eof) )
						YY_NEW_FILE;
#ifdef __cplusplus
					return yyinput();
#else
					return input();
#endif
					}

				case EOB_ACT_CONTINUE_SCAN:
					(yy_c_buf_p) = (yytext_ptr) + offset;
					break;
				}
			}
		}

	c = *(unsigned char *) (yy_c_buf_p);	/* cast for 8-bit char's */
	*(yy_c_buf_p) = '\0';	/* preserve cfgl_text */
	(yy_hold_char) = *++(yy_c_buf_p);

	if ( c == '\n' )
		   
    cfgl_lineno++;
;

	return c;
}
#endif	/* ifndef YY_NO_INPUT */

/** Immediately switch to a different input stream.
 * @param input_file A readable stream.
 * 
 * @note This function does not reset the start condition to @c INITIAL .
 */
    void cfgl_restart  (FILE * input_file )
{
    
	if ( ! YY_CURRENT_BUFFER ){
        cfgl_ensure_buffer_stack ();
		YY_CURRENT_BUFFER_LVALUE =
            cfgl__create_buffer(cfgl_in,YY_BUF_SIZE );
	}

	cfgl__init_buffer(YY_CURRENT_BUFFER,input_file );
	cfgl__load_buffer_state( );
}

/** Switch to a different input buffer.
 * @param new_buffer The new input buffer.
 * 
 */
    void cfgl__switch_to_buffer  (YY_BUFFER_STATE  new_buffer )
{
    
	/* TODO. We should be able to replace this entire function body
	 * with
	 *		cfgl_pop_buffer_state();
	 *		cfgl_push_buffer_state(new_buffer);
     */
	cfgl_ensure_buffer_stack ();
	if ( YY_CURRENT_BUFFER == new_buffer )
		return;

	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*(yy_c_buf_p) = (yy_hold_char);
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = (yy_c_buf_p);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars);
		}

	YY_CURRENT_BUFFER_LVALUE = new_buffer;
	cfgl__load_buffer_state( );

	/* We don't actually know whether we did this switch during
	 * EOF (cfgl_wrap()) processing, but the only time this flag
	 * is looked at is after cfgl_wrap() is called, so it's safe
	 * to go ahead and always set it.
	 */
	(yy_did_buffer_switch_on_eof) = 1;
}

static void cfgl__load_buffer_state  (void)
{
    	(yy_n_chars) = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
	(yytext_ptr) = (yy_c_buf_p) = YY_CURRENT_BUFFER_LVALUE->yy_buf_pos;
	cfgl_in = YY_CURRENT_BUFFER_LVALUE->yy_input_file;
	(yy_hold_char) = *(yy_c_buf_p);
}

/** Allocate and initialize an input buffer state.
 * @param file A readable stream.
 * @param size The character buffer size in bytes. When in doubt, use @c YY_BUF_SIZE.
 * 
 * @return the allocated buffer state.
 */
    YY_BUFFER_STATE cfgl__create_buffer  (FILE * file, int  size )
{
	YY_BUFFER_STATE b;
    
	b = (YY_BUFFER_STATE) cfgl_alloc(sizeof( struct yy_buffer_state )  );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in cfgl__create_buffer()" );

	b->yy_buf_size = size;

	/* yy_ch_buf has to be 2 characters longer than the size given because
	 * we need to put in 2 end-of-buffer characters.
	 */
	b->yy_ch_buf = (char *) cfgl_alloc(b->yy_buf_size + 2  );
	if ( ! b->yy_ch_buf )
		YY_FATAL_ERROR( "out of dynamic memory in cfgl__create_buffer()" );

	b->yy_is_our_buffer = 1;

	cfgl__init_buffer(b,file );

	return b;
}

/** Destroy the buffer.
 * @param b a buffer created with cfgl__create_buffer()
 * 
 */
    void cfgl__delete_buffer (YY_BUFFER_STATE  b )
{
    
	if ( ! b )
		return;

	if ( b == YY_CURRENT_BUFFER ) /* Not sure if we should pop here. */
		YY_CURRENT_BUFFER_LVALUE = (YY_BUFFER_STATE) 0;

	if ( b->yy_is_our_buffer )
		cfgl_free((void *) b->yy_ch_buf  );

	cfgl_free((void *) b  );
}

/* Initializes or reinitializes a buffer.
 * This function is sometimes called more than once on the same buffer,
 * such as during a cfgl_restart() or at EOF.
 */
    static void cfgl__init_buffer  (YY_BUFFER_STATE  b, FILE * file )

{
	int oerrno = errno;
    
	cfgl__flush_buffer(b );

	b->yy_input_file = file;
	b->yy_fill_buffer = 1;

    /* If b is the current buffer, then cfgl__init_buffer was _probably_
     * called from cfgl_restart() or through yy_get_next_buffer.
     * In that case, we don't want to reset the lineno or column.
     */
    if (b != YY_CURRENT_BUFFER){
        b->yy_bs_lineno = 1;
        b->yy_bs_column = 0;
    }

        b->yy_is_interactive = file ? (isatty( fileno(file) ) > 0) : 0;
    
	errno = oerrno;
}

/** Discard all buffered characters. On the next scan, YY_INPUT will be called.
 * @param b the buffer state to be flushed, usually @c YY_CURRENT_BUFFER.
 * 
 */
    void cfgl__flush_buffer (YY_BUFFER_STATE  b )
{
    	if ( ! b )
		return;

	b->yy_n_chars = 0;

	/* We always need two end-of-buffer characters.  The first causes
	 * a transition to the end-of-buffer state.  The second causes
	 * a jam in that state.
	 */
	b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
	b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

	b->yy_buf_pos = &b->yy_ch_buf[0];

	b->yy_at_bol = 1;
	b->yy_buffer_status = YY_BUFFER_NEW;

	if ( b == YY_CURRENT_BUFFER )
		cfgl__load_buffer_state( );
}

/** Pushes the new state onto the stack. The new state becomes
 *  the current state. This function will allocate the stack
 *  if necessary.
 *  @param new_buffer The new state.
 *  
 */
void cfgl_push_buffer_state (YY_BUFFER_STATE new_buffer )
{
    	if (new_buffer == NULL)
		return;

	cfgl_ensure_buffer_stack();

	/* This block is copied from cfgl__switch_to_buffer. */
	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*(yy_c_buf_p) = (yy_hold_char);
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = (yy_c_buf_p);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars);
		}

	/* Only push if top exists. Otherwise, replace top. */
	if (YY_CURRENT_BUFFER)
		(yy_buffer_stack_top)++;
	YY_CURRENT_BUFFER_LVALUE = new_buffer;

	/* copied from cfgl__switch_to_buffer. */
	cfgl__load_buffer_state( );
	(yy_did_buffer_switch_on_eof) = 1;
}

/** Removes and deletes the top of the stack, if present.
 *  The next element becomes the new top.
 *  
 */
void cfgl_pop_buffer_state (void)
{
    	if (!YY_CURRENT_BUFFER)
		return;

	cfgl__delete_buffer(YY_CURRENT_BUFFER );
	YY_CURRENT_BUFFER_LVALUE = NULL;
	if ((yy_buffer_stack_top) > 0)
		--(yy_buffer_stack_top);

	if (YY_CURRENT_BUFFER) {
		cfgl__load_buffer_state( );
		(yy_did_buffer_switch_on_eof) = 1;
	}
}

/* Allocates the stack if it does not exist.
 *  Guarantees space for at least one push.
 */
static void cfgl_ensure_buffer_stack (void)
{
	yy_size_t num_to_alloc;
    
	if (!(yy_buffer_stack)) {

		/* First allocation is just for 2 elements, since we don't know if this
		 * scanner will even need a stack. We use 2 instead of 1 to avoid an
		 * immediate realloc on the next call.
         */
		num_to_alloc = 1;
		(yy_buffer_stack) = (struct yy_buffer_state**)cfgl_alloc
								(num_to_alloc * sizeof(struct yy_buffer_state*)
								);
		if ( ! (yy_buffer_stack) )
			YY_FATAL_ERROR( "out of dynamic memory in cfgl_ensure_buffer_stack()" );
								  
		memset((yy_buffer_stack), 0, num_to_alloc * sizeof(struct yy_buffer_state*));
				
		(yy_buffer_stack_max) = num_to_alloc;
		(yy_buffer_stack_top) = 0;
		return;
	}

	if ((yy_buffer_stack_top) >= ((yy_buffer_stack_max)) - 1){

		/* Increase the buffer to prepare for a possible push. */
		int grow_size = 8 /* arbitrary grow size */;

		num_to_alloc = (yy_buffer_stack_max) + grow_size;
		(yy_buffer_stack) = (struct yy_buffer_state**)cfgl_realloc
								((yy_buffer_stack),
								num_to_alloc * sizeof(struct yy_buffer_state*)
								);
		if ( ! (yy_buffer_stack) )
			YY_FATAL_ERROR( "out of dynamic memory in cfgl_ensure_buffer_stack()" );

		/* zero only the new slots.*/
		memset((yy_buffer_stack) + (yy_buffer_stack_max), 0, grow_size * sizeof(struct yy_buffer_state*));
		(yy_buffer_stack_max) = num_to_alloc;
	}
}

/** Setup the input buffer state to scan directly from a user-specified character buffer.
 * @param base the character buffer
 * @param size the size in bytes of the character buffer
 * 
 * @return the newly allocated buffer state object. 
 */
YY_BUFFER_STATE cfgl__scan_buffer  (char * base, yy_size_t  size )
{
	YY_BUFFER_STATE b;
    
	if ( size < 2 ||
	     base[size-2] != YY_END_OF_BUFFER_CHAR ||
	     base[size-1] != YY_END_OF_BUFFER_CHAR )
		/* They forgot to leave room for the EOB's. */
		return 0;

	b = (YY_BUFFER_STATE) cfgl_alloc(sizeof( struct yy_buffer_state )  );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in cfgl__scan_buffer()" );

	b->yy_buf_size = size - 2;	/* "- 2" to take care of EOB's */
	b->yy_buf_pos = b->yy_ch_buf = base;
	b->yy_is_our_buffer = 0;
	b->yy_input_file = 0;
	b->yy_n_chars = b->yy_buf_size;
	b->yy_is_interactive = 0;
	b->yy_at_bol = 1;
	b->yy_fill_buffer = 0;
	b->yy_buffer_status = YY_BUFFER_NEW;

	cfgl__switch_to_buffer(b  );

	return b;
}

/** Setup the input buffer state to scan a string. The next call to cfgl_lex() will
 * scan from a @e copy of @a str.
 * @param yystr a NUL-terminated string to scan
 * 
 * @return the newly allocated buffer state object.
 * @note If you want to scan bytes that may contain NUL values, then use
 *       cfgl__scan_bytes() instead.
 */
YY_BUFFER_STATE cfgl__scan_string (yyconst char * yystr )
{
    
	return cfgl__scan_bytes(yystr,strlen(yystr) );
}

/** Setup the input buffer state to scan the given bytes. The next call to cfgl_lex() will
 * scan from a @e copy of @a bytes.
 * @param yybytes the byte buffer to scan
 * @param _yybytes_len the number of bytes in the buffer pointed to by @a bytes.
 * 
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE cfgl__scan_bytes  (yyconst char * yybytes, yy_size_t  _yybytes_len )
{
	YY_BUFFER_STATE b;
	char *buf;
	yy_size_t n;
	int i;
    
	/* Get memory for full buffer, including space for trailing EOB's. */
	n = _yybytes_len + 2;
	buf = (char *) cfgl_alloc(n  );
	if ( ! buf )
		YY_FATAL_ERROR( "out of dynamic memory in cfgl__scan_bytes()" );

	for ( i = 0; i < _yybytes_len; ++i )
		buf[i] = yybytes[i];

	buf[_yybytes_len] = buf[_yybytes_len+1] = YY_END_OF_BUFFER_CHAR;

	b = cfgl__scan_buffer(buf,n );
	if ( ! b )
		YY_FATAL_ERROR( "bad buffer in cfgl__scan_bytes()" );

	/* It's okay to grow etc. this buffer, and we should throw it
	 * away when we're done.
	 */
	b->yy_is_our_buffer = 1;

	return b;
}

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif

static void yy_fatal_error (yyconst char* msg )
{
    	(void) fprintf( stderr, "%s\n", msg );
	exit( YY_EXIT_FAILURE );
}

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up cfgl_text. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		cfgl_text[cfgl_leng] = (yy_hold_char); \
		(yy_c_buf_p) = cfgl_text + yyless_macro_arg; \
		(yy_hold_char) = *(yy_c_buf_p); \
		*(yy_c_buf_p) = '\0'; \
		cfgl_leng = yyless_macro_arg; \
		} \
	while ( 0 )

/* Accessor  methods (get/set functions) to struct members. */

/** Get the current line number.
 * 
 */
int cfgl_get_lineno  (void)
{
        
    return cfgl_lineno;
}

/** Get the input stream.
 * 
 */
FILE *cfgl_get_in  (void)
{
        return cfgl_in;
}

/** Get the output stream.
 * 
 */
FILE *cfgl_get_out  (void)
{
        return cfgl_out;
}

/** Get the length of the current token.
 * 
 */
yy_size_t cfgl_get_leng  (void)
{
        return cfgl_leng;
}

/** Get the current token.
 * 
 */

char *cfgl_get_text  (void)
{
        return cfgl_text;
}

/** Set the current line number.
 * @param line_number
 * 
 */
void cfgl_set_lineno (int  line_number )
{
    
    cfgl_lineno = line_number;
}

/** Set the input stream. This does not discard the current
 * input buffer.
 * @param in_str A readable stream.
 * 
 * @see cfgl__switch_to_buffer
 */
void cfgl_set_in (FILE *  in_str )
{
        cfgl_in = in_str ;
}

void cfgl_set_out (FILE *  out_str )
{
        cfgl_out = out_str ;
}

int cfgl_get_debug  (void)
{
        return cfgl__flex_debug;
}

void cfgl_set_debug (int  bdebug )
{
        cfgl__flex_debug = bdebug ;
}

static int yy_init_globals (void)
{
        /* Initialization is the same as for the non-reentrant scanner.
     * This function is called from cfgl_lex_destroy(), so don't allocate here.
     */

    /* We do not touch cfgl_lineno unless the option is enabled. */
    cfgl_lineno =  1;
    
    (yy_buffer_stack) = 0;
    (yy_buffer_stack_top) = 0;
    (yy_buffer_stack_max) = 0;
    (yy_c_buf_p) = (char *) 0;
    (yy_init) = 0;
    (yy_start) = 0;

/* Defined in main.c */
#ifdef YY_STDINIT
    cfgl_in = stdin;
    cfgl_out = stdout;
#else
    cfgl_in = (FILE *) 0;
    cfgl_out = (FILE *) 0;
#endif

    /* For future reference: Set errno on error, since we are called by
     * cfgl_lex_init()
     */
    return 0;
}

/* cfgl_lex_destroy is for both reentrant and non-reentrant scanners. */
int cfgl_lex_destroy  (void)
{
    
    /* Pop the buffer stack, destroying each element. */
	while(YY_CURRENT_BUFFER){
		cfgl__delete_buffer(YY_CURRENT_BUFFER  );
		YY_CURRENT_BUFFER_LVALUE = NULL;
		cfgl_pop_buffer_state();
	}

	/* Destroy the stack itself. */
	cfgl_free((yy_buffer_stack) );
	(yy_buffer_stack) = NULL;

    /* Reset the globals. This is important in a non-reentrant scanner so the next time
     * cfgl_lex() is called, initialization will occur. */
    yy_init_globals( );

    return 0;
}

/*
 * Internal utility routines.
 */

#ifndef yytext_ptr
static void yy_flex_strncpy (char* s1, yyconst char * s2, int n )
{
	register int i;
	for ( i = 0; i < n; ++i )
		s1[i] = s2[i];
}
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (yyconst char * s )
{
	register int n;
	for ( n = 0; s[n]; ++n )
		;

	return n;
}
#endif

void *cfgl_alloc (yy_size_t  size )
{
	return (void *) malloc( size );
}

void *cfgl_realloc  (void * ptr, yy_size_t  size )
{
	/* The cast to (char *) in the following accommodates both
	 * implementations that use char* generic pointers, and those
	 * that use void* generic pointers.  It works with the latter
	 * because both ANSI C and C++ allow castless assignment from
	 * any pointer type to void*, and deal with argument conversions
	 * as though doing an assignment.
	 */
	return (void *) realloc( (char *) ptr, size );
}

void cfgl_free (void * ptr )
{
	free( (char *) ptr );	/* see cfgl_realloc() for (char *) cast */
}

#define YYTABLES_NAME "yytables"

/*** THIS IS THE FLEX END ***/

static char* cfgl_parse_top(char *fn,void *ptr, cfgl_field_t *fields)
{
  FILE *f;
  f= fopen(fn,"r");
  if (!f) {
    snprintf(cfgl_ebuf,sizeof(cfgl_ebuf)-1, "can't open '%s'\n", fn); 
    return cfgl_ebuf;
  }
  cfgl_fname= fn;
  if (setjmp(cfgl_elabel)) 
     return cfgl_ebuf;
  yyin= f;
  yylex();
  while(token!='E') 
    cfgl_parse_field(ptr, fields);
  fclose(f);
  return 0;
}


static void cfgl_parse_struct(void *ptr, cfgl_field_t *fields)
{
  yylex();
  if (token=='{') {
    yylex();
    while(token!='}')
      cfgl_parse_field(ptr, fields);
    yylex();
  } else cfgl_parse_field(ptr, fields);
}


static void cfgl_parse_field(void *ptr, cfgl_field_t *fields)
{
  int i;
  if (token!='A') 
    cfgl_error("an identifier was expected");
  for(i=0;fields[i].name;i++)
   if (!strcmp(fields[i].name, yytext)) {
     cfgl_field_t *F= fields+i;
     switch(F->kind) {
     case 'I': cfgl_parse_intarray(ptr + F->offset,
                     ptr+ F->soffset, ptr+ F->noffset, F->arg); return;
     case 'F': cfgl_parse_floarray(ptr + F->offset,
                     ptr+ F->soffset, ptr+ F->noffset, F->arg); return;
     case 'S': cfgl_parse_strarray(ptr + F->offset,
                     ptr+ F->soffset, ptr+ F->noffset, F->arg); return;
     case 'R': cfgl_parse_recarray(ptr+F->offset, 
                    ptr+ F->soffset, ptr+F->noffset, F->eltsize,
                    F->arg); return;
     case 'i': cfgl_parse_int(ptr+F->offset, ptr+F->goffset, F->arg); return;
     case 'f': cfgl_parse_flo(ptr+F->offset, ptr+F->goffset, F->arg); return;
     case 's': cfgl_parse_str(ptr+F->offset, ptr+F->goffset, F->arg); return;
     case 'r': cfgl_parse_struct(ptr+F->offset, F->arg); return;
     }
   }
   cfgl_error("unknown field '%s'", yytext);
}


static void cfgl_parse_int(int *ptr, char *given, cfgl_intcst_t *cst)
{
  yylex();
  if (token=='I') { *ptr= ivalue; }
  else if (token=='F') { *ptr= fvalue; }
  else if (token=='A') { cfgl_lookup_int(cst,ptr); }
  else cfgl_error("syntax error. expecting an integer");
  yylex();
  *given= 1;
}

static void cfgl_lookup_int(cfgl_intcst_t *cst, int *ptr)
{
  int i;
  if (cst) for(i=0;cst[i].name;i++)
    if (!strcmp(cst[i].name, yytext)) { *ptr= cst[i].value; return ; }
  cst= cfgl_global_intcst;
  for(i=0;cst[i].name;i++)
    if (!strcmp(cst[i].name, yytext)) { *ptr= cst[i].value; return ; }
  cfgl_error("unknown constant '%s'", yytext);
}

static void cfgl_lookup_flo(cfgl_flocst_t *cst, double *ptr)
{
  int i;
  if (cst) for(i=0;cst[i].name;i++)
    if (!strcmp(cst[i].name, yytext)) { *ptr= cst[i].value; return ; }
  cst= cfgl_global_flocst;
  for(i=0;cst[i].name;i++)
    if (!strcmp(cst[i].name, yytext)) { *ptr= cst[i].value; return ; }
  cfgl_error("unknown constant '%s'", yytext);
}

static void cfgl_lookup_str(cfgl_strcst_t *cst, char **ptr)
{
  int i;
  if (cst) for(i=0;cst[i].name;i++)
    if (!strcmp(cst[i].name, yytext)) { *ptr= strdup(cst[i].value); return ; }
  cst= cfgl_global_strcst;
  for(i=0;cst[i].name;i++)
    if (!strcmp(cst[i].name, yytext)) { *ptr= strdup(cst[i].value); return ; }
  cfgl_error("unknown constant '%s'", yytext);
}

static void cfgl_parse_intarray
   (int **ptr, int *size, int *len, cfgl_intcst_t *cst)
{
  yylex();
  if (token=='{') {
    yylex();
    while(token!='}') 
      cfgl_put_int(ptr, size, len, cst);
    yylex();
  } else cfgl_put_int(ptr, size, len,cst);
}

static void cfgl_put_int
  (int **ptr, int *size, int *len, cfgl_intcst_t *cst)
{
  int *v;
  if (token=='}') return ;
  v= cfgl_array_grow(ptr, size, len, sizeof(int));
  if (token=='I')      { *v= ivalue; } 
  else if (token=='F') { *v= fvalue; }
  else if (token=='A') { cfgl_lookup_int(cst,v); }
  else cfgl_error("a number token or constant was expected");
  (*len)++;
  yylex();
}


static void* cfgl_array_grow(void *ptr, int *size, int *len, int esize)
{
  void **P= (void**) ptr;
  if (*len==*size) {
    if (!*size) {
      *size= 10;
      *P= calloc(1,*size * esize);
    } else {
      *size += *size/2;
      *P= realloc(*P, *size*esize);
      memset(*P + *len*esize, 0, (*size-*len)*esize);
    }
  }
  return (*P) + *len*esize;
}

static void cfgl_parse_recarray
    (void **ptr, int *size, int *len, int esize, cfgl_field_t *fields)
{
  void *P;
  yylex();
  P= cfgl_array_grow(ptr, size, len, esize);
  if (token=='{') {
    yylex();
    while(token!='}')
      cfgl_parse_field(P, fields);
    yylex();
  } else cfgl_parse_field(P, fields);
  (*len)++;
}


static void cfgl_parse_flo(double *ptr, char *given, cfgl_flocst_t *cst)
{
  yylex();
  if (token=='I') { *ptr= ivalue; }
  else if (token=='F') { *ptr= fvalue; }
  else if (token=='A') { cfgl_lookup_flo(cst,ptr); }
  else cfgl_error("syntax error. expecting an floating point number");
  *given= 1;
  yylex();
}

static void cfgl_parse_floarray
   (double **ptr, int *size, int *len, cfgl_flocst_t *cst)
{
  yylex();
  if (token=='{') {
    yylex();
    while(token!='}') 
      cfgl_put_flo(ptr, size, len, cst);
    yylex();
  } else cfgl_put_flo(ptr, size, len,cst);
}

static void cfgl_put_flo
  (double **ptr, int *size, int *len, cfgl_flocst_t *cst)
{
  double *v;
  if (token=='}') return ;
  v= cfgl_array_grow(ptr, size, len, sizeof(double));
  if (token=='I')      { *v= ivalue; } 
  else if (token=='F') { *v= fvalue; }
  else if (token=='A') { cfgl_lookup_flo(cst,v); }
  else cfgl_error("a number token or constant was expected");
  (*len)++;
  yylex();
}

static void cfgl_parse_str(char** ptr, char *given, cfgl_strcst_t *cst)
{
  yylex();
  if (token=='S') { *ptr= strdup(sbuf); }
  else if (token=='A') { cfgl_lookup_str(cst,ptr); }
  else cfgl_error("syntax error. expecting a string value");
  *given= 1;
  yylex();
}

static void cfgl_parse_strarray
   (char ***ptr, int *size, int *len, cfgl_strcst_t *cst)
{
  yylex();
  if (token=='{') {
    yylex();
    while(token!='}') 
      cfgl_put_str(ptr, size, len, cst);
    yylex();
  } else cfgl_put_str(ptr, size, len,cst);
}

static void cfgl_put_str
  (char ***ptr, int *size, int *len, cfgl_strcst_t *cst)
{
  char** v;
  if (token=='}') return ;
  v= cfgl_array_grow(ptr, size, len, sizeof(char*));
  if (token=='S')      { *v= strdup(sbuf); } 
  else if (token=='A') { cfgl_lookup_str(cst,v); }
  else cfgl_error("a string value or constant was expected");
  (*len)++;
  yylex();
}


static void cfgl_error(char *fmt, ...)
{
  char eb[2000];
  va_list ap;
  va_start(ap, fmt);
  vsnprintf(eb,2000,fmt,ap);
  va_end(ap);
  snprintf(cfgl_ebuf,2000, "error %s:%d:%s\n", cfgl_fname, yylineno, eb);
  longjmp(cfgl_elabel, 1);
}


static void sreset() { slen= 0; if (!sbuf) sbuf=malloc(ssize+1); }
static void sput(int x) { 
  if (slen==ssize) {
    ssize+= ssize/2;
    sbuf= realloc(sbuf, ssize+1);
  }
  sbuf[slen++]= x;
}

static char* cfgl_write_top(char *fn,void *ptr, cfgl_field_t *fields)
{
  FILE *f;
  f= fopen(fn,"w");
  if (!f) {
    snprintf(cfgl_ebuf,sizeof(cfgl_ebuf)-1, "can't open '%s'\n", fn); 
    return cfgl_ebuf;
  }
  cfgl_fname= fn;
  if (setjmp(cfgl_elabel)) 
     return cfgl_ebuf;

  
  while(fields->name)
  {
    cfgl_write_field(f, ptr, fields);
    fields++;
  }

  fclose(f);
  return NULL;
}

static void cfgl_write_field(FILE *out, void *ptr, cfgl_field_t *field)
{
  char *given;
  int  *N;
  char *dptr;

  given= ptr + field->goffset;
  N= (int*) (ptr+field->noffset);

    /* Note that the 'given' field isn't used for 'r' type entries.
       Each element of a record has its own 'given' field. We need
       to check those individually in order to decide whether an 
       r-type entry has any meaningful data and needs to be written. */
  switch(field->kind)
  {
  case 'i': case 'f': case 's': 
    if (!given[0]) return ;
    break;
  case 'I': case 'F': case 'S': case 'R':
    if (!N[0]) return;
    break;
  case 'r':
    if (has_valid_fields(ptr+field->offset, field->arg))
    {
       print_spaces(out);
       fprintf(out, "%s {\n", field->name);
       spaces+= 3;
       cfgl_write_struct(out, ptr+field->offset, field->arg);
       spaces-= 3;
       fprintf(out, "}\n");
    }
    return ;
  default:
     return ;
  }

  dptr= (char*) ptr + field->offset;
  switch(field->kind)
  {
  case 'i':
     print_spaces(out);
     fprintf(out, "%s %d\n", field->name, *(int*)dptr);
     return;
  case 'f':
     print_spaces(out);
     fprintf(out, "%s %f\n", field->name, *(double*)dptr);
     return;
  case 's':
     print_spaces(out);
     fprintf(out, "%s ", field->name);
     cfgl_write_str(out, *(void**)dptr);
     fprintf(out, "\n");
     return ;
  }

  int i; void *eptr;
  eptr= *(void**) dptr;

  for(i=0;i<N[0];i++)
  {
    print_spaces(out);
    switch(field->kind)
    {
    case 'I':
      fprintf(out, "%s %d\n", field->name, *(int*)eptr);
      eptr+= sizeof(int);
      break;
    case 'F':
      fprintf(out, "%s %f\n", field->name, *(double*)eptr);
      eptr+= sizeof(double);
      break;
    case 'S':
      fprintf(out,"%s ", field->name);
      cfgl_write_str(out, *(void**)dptr);
      fprintf(out, "\n");
      eptr+= sizeof(char*);
      break;
    case 'R':
      if (has_valid_fields(eptr, field->arg))
      {
        fprintf(out, "%s {\n", field->name);
        spaces+= 3;
        cfgl_write_struct(out, eptr, field->arg);
        spaces-= 3;
        fprintf(out, "}\n" );
      }
      eptr+= field->eltsize;
      break;
    }
  }
}

static int has_valid_fields(void *ptr, cfgl_field_t* field)
{
  int i;
  int N;
  void *eptr; 

  while(field->name)
  {
    switch(field->kind)
    {
      case 'i': case 'f': case 's':
        if (*(char*)(ptr + field->goffset))  return 1;
        break;
      case 'I': case 'F': case 'S': 
        if (*(int*)(ptr + field->noffset))  return 1;
        break;
      case 'r':
        if (has_valid_fields(ptr+field->offset, field->arg))
           return 1;
        break;
      case 'R':
        N= *(int*)(ptr+field->noffset);
        if (N==0) break;
        eptr= *(void**)(ptr+field->offset);
        for(i=0;i<N;i++)
        {
          if (has_valid_fields(eptr, field->arg)) return 1;
          eptr+= field->eltsize;
        }
        break;
    }
    field++;
  }
  return 0;
}

static void cfgl_write_struct(FILE *f, void *ptr, cfgl_field_t *field)
{
  while(field->name)
  {
    cfgl_write_field(f, ptr, field);
    field++;
  }
}

static void cfgl_write_str(FILE *f, char *str)
{
  if (!str) str= "";
  fprintf(f,"\"");
  while(*str)
  {
    unsigned char K= *str;
    if (K<32 || K>126) fprintf(f,"\\%03o", K);
                  else fprintf(f,"%c",K);
    str++;
  }
  fprintf(f,"\"");
}

static void print_spaces(FILE *out)
{
  int i;
  for(i=0;i<spaces;i++) fprintf(out," ");
}

char *cfg_parse_top(char *fn, cfg_top_t *obj)
{ return cfgl_parse_top(fn, obj, cfgl_T6); }
char *cfg_write_top(char *fn, cfg_top_t *obj)
{ return cfgl_write_top(fn, obj, cfgl_T6); }
