/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include <syslog.h>
#include <X11/Xlib.h>
#include "x-basics.h"
#include "x-props.h"
#include "config.h"
#include "bfont.h"
#include "panel.h"
#include "txtren.h"
#include "menu.h"
#include "graymap.h"
#include "medit.h"
#include "p-box.h"
#include "retarded.h"

static cfg_top_t cfg;

static char *appmenu_config_fn;

static void config_done(void *ignore)
{
  reload_start_menu(appmenu_config_fn);
  box_reload(0, cfg.left_box_file);
  box_reload(1, cfg.right_box_file);
  pager_resized();
  //printf("configuration is done, now all we need is to "
  //       "propagate it.\n");
}

void execute_program(char *x)
{
  char **argv;
  int status;

  while(waitpid(-1, &status, WNOHANG)>0);

  if (!x) return ;
  if (x[0]==':')
  {
    if (!strcmp(x,":configure"))
    {
      edit_app_menu(x_root_window, 300, 200,
                  appmenu_config_fn, &cfg,
                config_done, NULL);
      return ;
    }
    return ;
  }
  if (fork()) return ;
  argv= stupid_gnome_unquote_split(x);
  if (argv[0] && argv[0][0])
     execvp(argv[0], argv);
  exit(0);
}

static char *combinefn(char *A, char *B)
{
  char buf[2000];
  memset(buf,0, sizeof(buf));
  snprintf(buf, sizeof(buf), "%s/%s",A,B);
  buf[sizeof(buf)-1]= 0;
  return strdup(buf);
}

static void read_config()
{
  char *cfg_dir, *cfg_fn;
  char *xdg_config_home, *home;
  xdg_config_home= getenv("XDG_CONFIG_HOME");
  if (!xdg_config_home)
  {
    home= getenv("HOME");
    if (!home) return ;
    xdg_config_home= combinefn(home,".config");
  } else xdg_config_home= strdup(xdg_config_home);

  cfg_dir= combinefn(xdg_config_home, "mupanel");
  mkdir(xdg_config_home, 0700);
  mkdir(cfg_dir, 0700);
  free(xdg_config_home);

  cfg_fn= combinefn(cfg_dir,"config");
  cfg_parse_top(cfg_fn,&cfg);
  free(cfg_fn);
  
  appmenu_config_fn= combinefn(cfg_dir, "appmenu");

  cfg.left_box_file= combinefn(cfg_dir, "leftbox");
  cfg.right_box_file= combinefn(cfg_dir, "rightbox");
  free(cfg_dir);
}

static int x_error_handler(Display *disp, XErrorEvent *event)
{
static const struct { const char *name; int code; }
 xerrordef[]=
{
{"Unknown request",-1},
{"X_CreateWindow",1},
{"X_ChangeWindowAttributes",2},
{"X_GetWindowAttributes",3},
{"X_DestroyWindow",4},
{"X_DestroySubwindows",5},
{"X_ChangeSaveSet",6},
{"X_ReparentWindow",7},
{"X_MapWindow",8},
{"X_MapSubwindows",9},
{"X_UnmapWindow",10},
{"X_UnmapSubwindows",11},
{"X_ConfigureWindow",12},
{"X_CirculateWindow",13},
{"X_GetGeometry",14},
{"X_QueryTree",15},
{"X_InternAtom",16},
{"X_GetAtomName",17},
{"X_ChangeProperty",18},
{"X_DeleteProperty",19},
{"X_GetProperty",20},
{"X_ListProperties",21},
{"X_SetSelectionOwner",22},
{"X_GetSelectionOwner",23},
{"X_ConvertSelection",24},
{"X_SendEvent",25},
{"X_GrabPointer",26},
{"X_UngrabPointer",27},
{"X_GrabButton",28},
{"X_UngrabButton",29},
{"X_ChangeActivePointerGrab",30},
{"X_GrabKeyboard",31},
{"X_UngrabKeyboard",32},
{"X_GrabKey",33},
{"X_UngrabKey",34},
{"X_AllowEvents",35},
{"X_GrabServer",36},
{"X_UngrabServer",37},
{"X_QueryPointer",38},
{"X_GetMotionEvents",39},
{"X_TranslateCoords",40},
{"X_WarpPointer",41},
{"X_SetInputFocus",42},
{"X_GetInputFocus",43},
{"X_QueryKeymap",44},
{"X_OpenFont",45},
{"X_CloseFont",46},
{"X_QueryFont",47},
{"X_QueryTextExtents",48},
{"X_ListFonts",49},
{"X_ListFontsWithInfo",50},
{"X_SetFontPath",51},
{"X_GetFontPath",52},
{"X_CreatePixmap",53},
{"X_FreePixmap",54},
{"X_CreateGC",55},
{"X_ChangeGC",56},
{"X_CopyGC",57},
{"X_SetDashes",58},
{"X_SetClipRectangles",59},
{"X_FreeGC",60},
{"X_ClearArea",61},
{"X_CopyArea",62},
{"X_CopyPlane",63},
{"X_PolyPoint",64},
{"X_PolyLine",65},
{"X_PolySegment",66},
{"X_PolyRectangle",67},
{"X_PolyArc",68},
{"X_FillPoly",69},
{"X_PolyFillRectangle",70},
{"X_PolyFillArc",71},
{"X_PutImage",72},
{"X_GetImage",73},
{"X_PolyText8",74},
{"X_PolyText16",75},
{"X_ImageText8",76},
{"X_ImageText16",77},
{"X_CreateColormap",78},
{"X_FreeColormap",79},
{"X_CopyColormapAndFree",80},
{"X_InstallColormap",81},
{"X_UninstallColormap",82},
{"X_ListInstalledColormaps",83},
{"X_AllocColor",84},
{"X_AllocNamedColor",85},
{"X_AllocColorCells",86},
{"X_AllocColorPlanes",87},
{"X_FreeColors",88},
{"X_StoreColors",89},
{"X_StoreNamedColor",90},
{"X_QueryColors",91},
{"X_LookupColor",92},
{"X_CreateCursor",93},
{"X_CreateGlyphCursor",94},
{"X_FreeCursor",95},
{"X_RecolorCursor",96},
{"X_QueryBestSize",97},
{"X_QueryExtension",98},
{"X_ListExtensions",99},
{"X_ChangeKeyboardMapping",100},
{"X_GetKeyboardMapping",101},
{"X_ChangeKeyboardControl",102},
{"X_GetKeyboardControl",103},
{"X_Bell",104},
{"X_ChangePointerControl",105},
{"X_GetPointerControl",106},
{"X_SetScreenSaver",107},
{"X_GetScreenSaver",108},
{"X_ChangeHosts",109},
{"X_ListHosts",110},
{"X_SetAccessControl",111},
{"X_SetCloseDownMode",112},
{"X_KillClient",113},
{"X_RotateProperties",114},
{"X_ForceScreenSaver",115},
{"X_SetPointerMapping",116},
{"X_GetPointerMapping",117},
{"X_SetModifierMapping",118},
{"X_GetModifierMapping",119},
{"X_NoOperation",127},
{NULL, 0}};
   char g[2000];
   int i;
   for(i=1;xerrordef[i].name;i++)
      if (xerrordef[i].code==event->request_code) break;
   if (!xerrordef[i].name) i= 0;
   memset(g,0,sizeof(g));
   XGetErrorText(disp, event->error_code,g, sizeof(g)-1);
   syslog(LOG_NOTICE,"X11 error: [%d] %s. Resource: %x. Request: %s[%d.%d]",
                       event->error_code, g, 
                       (unsigned int) event->resourceid, 
                       xerrordef[i].name,
                       event->request_code, event->minor_code);
   return 1;
}



int main()
{
  if (init_display(NULL)) return 1;
  openlog("mupanel", 0, LOG_USER); 
  XSetErrorHandler(x_error_handler);
  read_config();
  x_atoms_init();
  font_init();
  init_builtin_font();
  menu_init(&cfg.menu_style);
  graymap_init(x_red_shift,x_green_shift,x_blue_shift);
  init_panel(appmenu_config_fn, &cfg);
  panel_loop();
  closelog();
  return 0;
}
