/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-image.h"
#include "config.h"
#include "menu.h"
#include "imgload.h"
#include "txtren.h"
#include "appmenu.h"
#include "panel.h"
#include "graymap.h"
#include "x-events.h"
#include "p-box.h"


#define MAX_BUT 40

typedef struct {
  ximage_t *image;
  Window window;
  int n_ico;
  char *str[MAX_BUT];
  int x[MAX_BUT],w[MAX_BUT];
  panel_part_t *part;
} box_t;

static box_t boxes[2];

static int    BO_button_size;
static int    BO_pad;
static int    BO_bg;
static Window BO_parent;
static int    BO_parent_height;

static void box_expose(box_t *B,XExposeEvent *ev);
static void box_click(box_t *B,XButtonEvent *ev);

static void box_destroy(int no);


void init_box(Window parent, int parent_h,
              int no,
              char *box_file_name,
              int button_size, int pad, uint32_t bg,
              panel_part_t *part)
{
  boxes[no].part= part;
  BO_parent= parent;
  BO_parent_height= parent_h;
  BO_button_size= button_size;
  BO_pad= pad;
  BO_bg= bg;

  if (box_file_name) 
     box_reload(no, box_file_name);
  else
     box_destroy(no);
}

static void box_click(box_t *B,XButtonEvent *ev)
{
  int i;
  for(i=0;i<B->n_ico;i++)
   if (ev->x>=B->x[i] && ev->x<B->x[i]+B->w[i])
   {
     execute_program(B->str[i]);
     break;
   }
}

static void box_expose(box_t *B,XExposeEvent *ev)
{
  put_image(B->window, ev->x, ev->y,
            B->image, ev->x, ev->y, 
            ev->width, ev->height);
}

void box_reload(int no, char *fn)
{
  amenu_t *buttons;
  int i, tw,bx;
  uint32_t iW[MAX_BUT], iH[MAX_BUT], *iD[MAX_BUT], def[MAX_BUT];
  Window win;
  int n_buttons;
  aitem_t *it;
  box_t *B;

  box_destroy(no);

  B= boxes+no;

  buttons= read_app_menu(fn,NULL);
  if (buttons->items.next==&buttons->items) goto empty;

  i= 0;
  for(it=buttons->items.next;it!=&buttons->items && i<MAX_BUT;it=it->next)
  {
     if (it->type!=0) continue;
     def[i]= 0;
     if (it->icon && load_icon(it->icon,BO_button_size, BO_button_size,
                                        &iW[i], &iH[i], &iD[i]))
     {
       load_icon(":default-button.png", BO_button_size, BO_button_size,
                                        &iW[i], &iH[i], &iD[i]);
       def[i]= 1;
     }
     B->str[i]= strdup(it->exec);
     i++;
  }
  n_buttons= i;
  if (n_buttons==0) goto empty;

  tw= - BO_pad;
  for(i=0;i<n_buttons;i++)
    tw+= iW[i] + BO_pad;

  B->image= create_image( tw, BO_parent_height);
  fill_rect(B->image->data, tw, BO_parent_height,
              0,0, tw, BO_parent_height, BO_bg);

  bx= 0;
  for(i=0;i<n_buttons;i++)
  {
    blend_alpha_image(B->image->data, tw, BO_parent_height,
                         bx, (BO_parent_height-iH[i])/2,
                         iD[i], iW[i], iH[i], 0, 0, iW[i], iH[i], BO_bg);
    B->x[i]= bx;
    B->w[i]= iW[i];
    if (!def[i]) free(iD[i]);
    bx+= iW[i] + BO_pad;
  }
  B->n_ico= i;

  win= create_window_Ob(BO_parent, 0,0,
                                     tw, BO_parent_height,
                                     ButtonPressMask | ExposureMask,0);
  B->part->window= B->window= win;
  B->part->w= tw;

  register_event_handler(win, Expose, B, box_expose, NULL);
  register_event_handler(win, ButtonPress, B, box_click, NULL);
empty:
  destroy_app_menu(buttons);
}

static void box_destroy(int no)
{
  int i;
  box_t *B;
  B= boxes+no;
  if (B->window!=None)
  {
    unregister_event_handler(B->window, Expose, NULL);
    unregister_event_handler(B->window, ButtonPress, NULL);
    XDestroyWindow(x_display, B->window);
    B->window= B->part->window= None;
  }
  for(i=0;i<B->n_ico;i++)
    free(B->str[i]);
  B->n_ico= 0;
  if (B->image) destroy_image(B->image);
  B->image= NULL;
}
