/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#include <X11/Xlib.h>
#include <stdint.h>
#include "x-basics.h"
#include "x-events.h"



typedef struct 
  { Window window; unsigned int event;
    void *data; void (*func)(void *, XEvent *); }
  event_handler_t;

static int n_event_handlers, s_event_handlers;
static event_handler_t *event_handlers;

static event_handler_t *find_event(Window window, unsigned int event);

void register_event_handler(Window window, unsigned int event,
                         void *data, void *func, void**old)
{
  event_handler_t *EV,*CP;
  EV= find_event(window, event);
  if (old) *old= NULL;
  if (EV)
  {
    if (old)
    {
      CP= malloc(sizeof(*CP));
      *CP= *EV;
      *old= CP;
    }
    goto done;
  }

  if (n_event_handlers==s_event_handlers)
  {
    if (s_event_handlers==0)
    {
      s_event_handlers= 20;
      event_handlers= malloc(sizeof(event_handlers[0])*s_event_handlers);
    }
    else
    {
      s_event_handlers+= s_event_handlers/2;
      event_handlers= realloc(event_handlers,
                              sizeof(event_handlers[0])*s_event_handlers);
    }
  }
  EV= event_handlers+n_event_handlers;
  n_event_handlers++;

done:
  EV->window= window;
  EV->event= event;
  EV->data= data;
  EV->func= func;
}

void unregister_all_event_handlers(Window window)
{
  int i;
  for(i=0;i<n_event_handlers;i++)
    if (event_handlers[i].window==window)
    {
      event_handlers[i]= event_handlers[n_event_handlers-1];
      n_event_handlers--; i--;
    }
}

void unregister_event_handler(Window window, unsigned int event, void *old)
{
  event_handler_t *EV,*NE;
  EV= old;
  NE= find_event(window, event);
  if (NE && EV)         *NE= *EV; 
  else if (NE && !EV)   *NE= event_handlers[--n_event_handlers];
  else if (!NE && EV)   register_event_handler(window, event,
                                               EV->data, EV->func,NULL);
  if (EV) free(EV);
}


static event_handler_t *find_event(Window window, unsigned int event)
{
  event_handler_t *EV;
  int i;
  for(i=0;i<n_event_handlers;i++)
  {
    EV= event_handlers+i;
    if (EV->window==window && EV->event==event)
       return EV;
  }
  return NULL;
}

#include <syslog.h>

void handle_registered_event(XEvent *event)
{
  event_handler_t *EV;

  if (event->xany.window==x_root_window)
  {  if (event->type==ButtonPress) syslog(LOG_USER|LOG_NOTICE,"root press\n");
    if (event->type==ButtonRelease) syslog(LOG_USER|LOG_NOTICE,"root release\n");
 }
                              

  EV= find_event(event->xany.window, event->type);
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void call_event_handler(void *H,XEvent *event)
{
  event_handler_t *EV;
  EV= H;
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void free_event_handler(void *H)
{
  if (H) free(H);
}


